#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <grp.h>
#include <list>
#include <OpenIPMI/ipmiif.h>
#include <OpenIPMI/ipmi_smi.h>
#include <OpenIPMI/ipmi_err.h>
#include <OpenIPMI/ipmi_auth.h>
#include <OpenIPMI/ipmi_lan.h>
#include <OpenIPMI/ipmi_posix.h>

using namespace std;

bool cluster::init(char *chassisfilename, char *clusterfilename, char *commandfile) {   
  unlink(nodewatchdat);
  os_hnd = ipmi_posix_setup_os_handler();
  os_hnd->set_log_handler(os_hnd, my_vlog);
  ipmi_init(os_hnd);

  parsechassis(chassisfilename);
  validatechassis();
  parsecluster(clusterfilename);

  setupcommandfifo(commandfile);
  if (processalerts==true) {
    setupemail();
  } else {
    error_msg((char *)"Skipping setupemail since alerts processing is not enabled in the cluster config file\n");
  }
  outputautoconf();
  lastoutputtime=time(NULL);    //set output time as current, so no output file for OUTPUTFREQ seconds after this
  return true;
}

void cluster::setupcommandfifo(char *filename) {
  unlink(filename);
  if (mkfifo(filename,S_IWUSR | S_IRUSR)!=0) error_msg((char *)"Failed to generate command fifo\n");
  commandfifo=open(filename,O_RDONLY|O_NONBLOCK);
  if(commandfifo==-1) error_msg((char *)"Failed to open command fifo\n");
}

void cluster::setupemail(){
  bool error=false;
  if(emailto[0]=='\0') {
    error_msg((char *) "Did not read email to address in cluster config file\n");
    error=true;
  }
  if(emailfrom[0]=='\0') {
    error_msg((char *) "Did not read email from address in cluster config file\n");
    error=true;
  }
  if(emailserver[0]=='\0') {
    error_msg((char *) "Did not read email server in cluster config file\n");
    error=true;
  }
  if (error==true) exit(1);
 
  pendingalerts.setupemail(emailto,emailfrom,emailserver);
  return;
}

void cluster::validatechassis(){
  int count,count2;
  for(count=0;count<numchassis; count++){
    for(count2=0;count2<mychassis[count].numfans;count2++){
      if(mychassis[count].fanname[count2]==NULL) { //if it equals null it was never set, so there was an error
        fprintf(stderr,"Failed to find or parse FAN%i for chassis %s\nExiting now.\n",count2+1,mychassis[count].name);
        exit(1);
      }
    }
    for(count2=0;count2<mychassis[count].numtemps;count2++){
      if(mychassis[count].tempname[count2]==NULL) { //if it equals null it was never set, so there was an error
        fprintf(stderr,"Failed to find or parse TEMP%i for chassis %s\nExiting now.\n",count2+1,mychassis[count].name);
        exit(1);
      }
    }
    for(count2=0;count2<mychassis[count].numvolts;count2++){
      if(mychassis[count].voltname[count2]==NULL) { //if it equals null it was never set, so there was an error
        fprintf(stderr,"Failed to find or parse VOLT%i for chassis %s\nExiting now.\n",count2+1,mychassis[count].name);
        exit(1);
      }
    }
  }
}

void cluster::parsechassis(char *chassisfilename){
  FILE *chassisconf;
  char line[120], *lineparse;
  rwwsensor tempsensor;
  int tempint;
  numchassis=0;

  chassisconf=fopen(chassisfilename,"r");
  if (chassisconf==NULL) {
     fprintf(stderr,"Unable to open chassis configuration file: %s\nExiting now.\n",chassisfilename);
     exit(1);
  }

  fgets(line,120,chassisconf); //start it with the first line :)
  while (!feof(chassisconf) && line!= NULL) {
    if (*line=='!' || *line=='#' || *line==' ' || *line=='\n' ) {  //ignore comment lines and blank lines
      fgets(line,120,chassisconf);
      continue;
    }
    if (strncmp(line,"$CHASSISTYPE",12)==0) { //marks the start of a new chassis type
      numchassis++;
      strncpy(mychassis[numchassis-1].name,&line[13],MAXCHASSISNAME); //FIXME- check if it already exists
      mychassis[numchassis-1].name[strlen(mychassis[numchassis-1].name)-1]='\0'; //change newline to EOS
      fgets(line,120,chassisconf);
      while( (!feof(chassisconf)) && ( (line[0]=='!') || (line[0]=='#') || (line[1]=='$') ) ) { //read in this chassis only
        if (*line=='!' || *line=='#' || *line==' ' || *line=='\n' ) {  //ignore comment lines and blank lines
          fgets(line,120,chassisconf);
          continue;
        }
        if (strncmp(line,"$$FANCOUNT",10)==0) {
          mychassis[numchassis-1].numfans=atoi(&line[11]);
        } else if (strncmp(line,"$$TEMPCOUNT",11)==0) {
          mychassis[numchassis-1].numtemps=atoi(&line[12]);
        } else if (strncmp(line,"$$VOLTCOUNT",11)==0) {
          mychassis[numchassis-1].numvolts=atoi(&line[12]);
        }
        //SENSORCOUNT (FAN, TEMP, VOLT) must be listed in configuration file first or this will fail
        else if( (mychassis[numchassis-1].numfans>0) && (strncmp(line,"$$FAN",5)==0) ) {
          if(line[6]==' ') { //single digit fan #
            line[6]='\0';
            lineparse=&line[5];
            tempint=atoi(lineparse)-1;
            lineparse=&line[7]; //remainder of the line after the FAN#
          } else { //double digit fan #
            line[7]='\0';
            lineparse=&line[5];
            tempint=atoi(lineparse)-1;
            lineparse=&line[8];
          }
          parsesensor(&tempsensor,lineparse);  //let the function extract short name ,long name,crit/alert
          mychassis[numchassis-1].fanshortname[tempint]=(char *)malloc(MAXSENSORNAME);
          mychassis[numchassis-1].fanname[tempint]=(char *)malloc(MAXSENSORNAME);
          strncpy(mychassis[numchassis-1].fanshortname[tempint],tempsensor.shortname,MAXSENSORNAME);
          strncpy(mychassis[numchassis-1].fanname[tempint],tempsensor.name,MAXSENSORNAME);
          mychassis[numchassis-1].fanlimit[tempint][0]=tempsensor.critlow;
          mychassis[numchassis-1].fanlimit[tempint][1]=tempsensor.alertlow;
          mychassis[numchassis-1].fanlimit[tempint][2]=tempsensor.alerthigh;
          mychassis[numchassis-1].fanlimit[tempint][3]=tempsensor.crithigh;
        }else if( (mychassis[numchassis-1].numtemps>0) && (strncmp(line,"$$TEMP",6)==0)) { 
          if(line[7]==' ') { //single digit temp #
            line[7]='\0';
            lineparse=&line[6];
            tempint=atoi(lineparse)-1;
            lineparse=&line[8]; //remainder of the line after the TEMP#
          } else { //double digit temp #
            line[8]='\0';
            lineparse=&line[6];
            tempint=atoi(lineparse)-1;
            lineparse=&line[9];
          }
          parsesensor(&tempsensor,lineparse);  //let the function extract short name ,long name,crit/alert
          mychassis[numchassis-1].tempshortname[tempint]=(char *)malloc(MAXSENSORNAME);
          mychassis[numchassis-1].tempname[tempint]=(char *)malloc(MAXSENSORNAME);
          strncpy(mychassis[numchassis-1].tempshortname[tempint],tempsensor.shortname,MAXSENSORNAME);
          strncpy(mychassis[numchassis-1].tempname[tempint],tempsensor.name,MAXSENSORNAME);
          mychassis[numchassis-1].templimit[tempint][0]=tempsensor.critlow;
          mychassis[numchassis-1].templimit[tempint][1]=tempsensor.alertlow;
          mychassis[numchassis-1].templimit[tempint][2]=tempsensor.alerthigh;
          mychassis[numchassis-1].templimit[tempint][3]=tempsensor.crithigh;
        } else if((mychassis[numchassis-1].numvolts>0) && (strncmp(line,"$$VOLT",6)==0)) {
          if(line[7]==' ') { //single digit volt #
            line[7]='\0';
            lineparse=&line[6];
            tempint=atoi(lineparse)-1;
            lineparse=&line[8]; //remainder of the line after the VOLT#
          } else { //double digit volt #
            line[8]='\0';
            lineparse=&line[6];
            tempint=atoi(lineparse)-1;
            lineparse=&line[9];
          }
          parsesensor(&tempsensor,lineparse);  //let the function extract short name ,long name,crit/alert
          mychassis[numchassis-1].voltshortname[tempint]=(char *)malloc(MAXSENSORNAME);
          mychassis[numchassis-1].voltname[tempint]=(char *)malloc(MAXSENSORNAME);
          strncpy(mychassis[numchassis-1].voltshortname[tempint],tempsensor.shortname,MAXSENSORNAME);
          strncpy(mychassis[numchassis-1].voltname[tempint],tempsensor.name,MAXSENSORNAME);
          mychassis[numchassis-1].voltlimit[tempint][0]=tempsensor.critlow;
          mychassis[numchassis-1].voltlimit[tempint][1]=tempsensor.alertlow;
          mychassis[numchassis-1].voltlimit[tempint][2]=tempsensor.alerthigh;
          mychassis[numchassis-1].voltlimit[tempint][3]=tempsensor.crithigh;
        } 
        if (!feof(chassisconf)) {
          memset(&line,'\0',119);
          fgets(line,120,chassisconf); //get the next line of this chassis definition
        }
        else break;
      }
    } else {
      fprintf(stderr,"Unknown line encountered in chassis configuration file:\n%s\nExiting now.\n",line);
    }
    if (!feof(chassisconf)) {
      memset(&line,'\0',119);
      fgets(line,120,chassisconf); //get the next line of this chassis definition
    }
    else break;
  }
}

void cluster::parsecluster(char *clusterfilename){
  FILE *clusterconf;
  char line[160];
  char *nodename, *chassisname, *username, *password, *emailsetting;
  char *ipmitarget[1],*ipmiport[1];
  int rv;

  numnodes=0;
  nodename=(char *)malloc(MAXNODENAME);
  chassisname=(char *)malloc(MAXCHASSISNAME);
  username=(char *)malloc(MAXNODENAME);
  password=(char *)malloc(MAXNODENAME);
  emailsetting=(char *)malloc(MAXEMAILADDRESS);
  ipmitarget[0]=(char *)malloc(MAXNODENAME);
  ipmiport[0]=(char *)malloc(5);

  strcpy(ipmiport[0],IPMIPORT);

  clusterconf=fopen(clusterfilename,"r");
  if (clusterconf==NULL) {
     fprintf(stderr,"Unable to open cluster configuration file: %s\nExiting now.\n",clusterfilename);
     exit(1);
  }

  fgets(line,120,clusterconf); //start it with the first line :)
  while (!feof(clusterconf) && line!= NULL) {
    if (*line=='!' || *line=='#' || *line==' ' || *line=='\n' ) {  //ignore comment lines and blank lines
      fgets(line,120,clusterconf);
      continue;
    }
    if (strncmp(line,"$NODE",5)==0) { //marks the start of a node
      numnodes++;
      parsenode(nodename,chassisname,username,password,&line[6]);
      mynodes[numnodes-1].thischassis=findchassis(chassisname);
      if(mynodes[numnodes-1].thischassis==NULL) {  //could not find that chassis
        fprintf(stderr,"Unable to locate chassis %s for node %s\nExiting now.\n",chassisname,nodename);
        exit(1);
      }
      mynodes[numnodes-1].name=(char *)malloc(strlen(nodename));
      mynodes[numnodes-1].username=(char *)malloc(strlen(username));
      mynodes[numnodes-1].password=(char *)malloc(strlen(password));
      strcpy(mynodes[numnodes-1].name,nodename);
      strcpy(mynodes[numnodes-1].username,username);
      strcpy(mynodes[numnodes-1].password,password);
       
//create ipmi connections and open domains here
      sprintf(ipmitarget[0],"%s-ipmi",nodename);
      
      rv = ipmi_ip_setup_con(ipmitarget,ipmiport,1,IPMI_AUTHTYPE_MD5,IPMI_PRIVILEGE_ADMIN,
                             username,(strlen(username)),password,(strlen(password)),
                             os_hnd,NULL,&mynodes[numnodes-1].ipmicon);
      if (rv) {
        fprintf(stderr, "ipmi_ip_setup_con: %s", strerror(rv));
        exit(1);
      }
      rv = ipmi_open_domain(nodename, &mynodes[numnodes-1].ipmicon, 1, domain_open,
                            NULL, domain_up, NULL, NULL, 0, NULL);
      if (rv) {
        fprintf(stderr, "ipmi_init_domain: %s\n", strerror(rv));
        exit(1);
      }

    } else if (strncmp(line,"$ENABLE_ALERTS",14)==0) {
      parseemail(emailsetting, &line[15]);
      if (strncasecmp("yes",emailsetting,3)==0) processalerts=true;
    } else if (strncmp(line,"$EMAIL_TO",9)==0) {
      parseemail(emailsetting, &line[10]);
      strncpy(emailto,emailsetting,MAXEMAILADDRESS);
    } else if (strncmp(line,"$EMAIL_FROM",11)==0) {
      parseemail(emailsetting, &line[12]);
      strncpy(emailfrom,emailsetting,MAXEMAILADDRESS);
    } else if (strncmp(line,"$EMAIL_SERVER",13)==0) {
      parseemail(emailsetting, &line[14]);
      strncpy(emailserver,emailsetting,MAXEMAILSERVER);
    } else if (strncmp(line,"$APACHE_GROUP",13)==0) {
      group *mygroupinfo;
      parseemail(emailsetting, &line[14]);
      mygroupinfo=getgrnam(emailsetting);  //yes.. emailsetting is a bad variable name, just reusing it
      apachegid=mygroupinfo->gr_gid;
    }      
    if (!feof(clusterconf)) fgets(line,120,clusterconf); //get the next line of this chassis definition
    else break;
  }
  free(nodename);
  free(chassisname);
  free(username);
  free(password);
  free(emailsetting);

  return;
}

bool cluster::command(char *name, int command){  //send a command to node "name"
  node *tempnode;
  tempnode=findnode(name);
  if (tempnode==NULL) {
    char temp[200];
    sprintf(temp,"Unable to locate node %s to issue command code %i.", name, command);
    error_msg(temp);
    return false;
  }
  return tempnode->command(command);
}

void cluster::process() {
  time_t newtime;
  getdata();
  if (processalerts==true) {
    checklimits();
    pendingalerts.process();
  }
  newtime=time(NULL);
  if ((newtime>lastoutputtime+OUTPUTFREQ) && domainisup) {
    outputdata();
    lastoutputtime=newtime;
  }
  processcommandfifo();
}

void cluster::processcommandfifo() {
  char buffer[8192], line[1024];
  int bytesread, index=0, lineindex=0;

  bytesread=read(commandfifo, buffer, 8192);
  if (bytesread!=0) {  //commands waiting, parse them!
    if (bytesread==8192) {
      error_msg((char *)"Command queue overflow.  Exitting.\n");
      exit(3);
    }
    while ( index < bytesread ) {
      lineindex=0;
      while ( buffer[index]!='\n') {
        line[lineindex]=buffer[index];
        index++;
        lineindex++;
      }
      index++;
      line[lineindex]='\0';

      if (strncmp(line,"$DO_POWER_ON",12)==0) { // power on node
        command(&line[13],POWERON);
      } else if (strncmp(line,"$DO_POWER_OFF",13)==0) {
        command(&line[14],POWEROFF);
      } else if (strncmp(line,"$DO_RESET",9)==0) {
        command(&line[10],RESET);
      } 
    }
  }
} 
  
void cluster::outputautoconf() {
  int count, count2;
  chassis *prevchassis;
  FILE *outputfile;
  char tempstring[4000];
  char fanlow[100],fancritlow[100],fanhigh[100],fancrithigh[100];
  char templow[100],tempcritlow[100],temphigh[100],tempcrithigh[100];
  char voltlow[100],voltcritlow[100],volthigh[100],voltcrithigh[100];
  mode_t oldumask;

  oldumask=umask(S_IWGRP | S_IWOTH | S_IROTH | S_IXOTH | S_IXGRP | S_IXUSR);
  unlink(autonodeconfdat);
  outputfile=fopen(autonodeconfdat,"w");
  if(outputfile==NULL) {
    fprintf(stderr,"Could not open %s for writing.\n",autonodeconfdat);
    exit(1);
  }
  prevchassis=NULL;
  
  for (count=0; count<numnodes ; count++){
    if ( prevchassis != mynodes[count].thischassis ) {  //mark chassis changes
      prevchassis=mynodes[count].thischassis;
      if (count==0) fprintf(outputfile,"$NEWCHASSIS");
      else fprintf(outputfile,"\n$NEWCHASSIS");
    }
    strcpy(fanlow,""); strcpy(fancritlow,""); strcpy(fanhigh,""); strcpy(fancrithigh,"");
    strcpy(templow,""); strcpy(tempcritlow,""); strcpy(temphigh,""); strcpy(tempcrithigh,"");
    strcpy(voltlow,""); strcpy(voltcritlow,""); strcpy(volthigh,""); strcpy(voltcrithigh,"");

    sprintf(tempstring,"\n$NODE %i %s\n$$CHASSIS %s\n$$USER %s\n$$PASSWORD %s\n"
            "$$SERIALCONSOLE\n$$FANLABELS ",
            count+1,mynodes[count].name,mynodes[count].thischassis->name,
            mynodes[count].username,mynodes[count].password);

    for (count2=0; count2<mynodes[count].thischassis->numfans; count2++) {
      sprintf(tempstring,"%s %s", tempstring, mynodes[count].thischassis->fanshortname[count2]);
      sprintf(fancritlow,"%s %i", fancritlow, (int)mynodes[count].thischassis->fanlimit[count2][0]);
      sprintf(fanlow,"%s %i", fanlow, (int)mynodes[count].thischassis->fanlimit[count2][1]);
      sprintf(fanhigh,"%s %i", fanhigh, (int)mynodes[count].thischassis->fanlimit[count2][2]);
      sprintf(fancrithigh,"%s %i", fancrithigh, (int)mynodes[count].thischassis->fanlimit[count2][3]);
    }
    sprintf(tempstring,"%s\n$$TEMPLABELS ", tempstring);
    for (count2=0; count2<mynodes[count].thischassis->numtemps; count2++) {
      sprintf(tempstring,"%s %s", tempstring, mynodes[count].thischassis->tempshortname[count2]);
      sprintf(tempcritlow,"%s %i", tempcritlow, (int)(mynodes[count].thischassis->templimit[count2][0]*100));
      sprintf(templow,"%s %i", templow, (int)(mynodes[count].thischassis->templimit[count2][1]*100));
      sprintf(temphigh,"%s %i", temphigh, (int)(mynodes[count].thischassis->templimit[count2][2]*100));
      sprintf(tempcrithigh,"%s %i", tempcrithigh, (int)(mynodes[count].thischassis->templimit[count2][3]*100));
    }
    sprintf(tempstring,"%s\n$$VOLTLABELS ", tempstring);
    for (count2=0; count2<mynodes[count].thischassis->numvolts; count2++) {
      sprintf(tempstring,"%s %s", tempstring, mynodes[count].thischassis->voltshortname[count2]);
      sprintf(voltcritlow,"%s %i", voltcritlow, (int)(mynodes[count].thischassis->voltlimit[count2][0]*100));
      sprintf(voltlow,"%s %i", voltlow, (int)(mynodes[count].thischassis->voltlimit[count2][1]*100));
      sprintf(volthigh,"%s %i", volthigh, (int)(mynodes[count].thischassis->voltlimit[count2][2]*100));
      sprintf(voltcrithigh,"%s %i", voltcrithigh, (int)(mynodes[count].thischassis->voltlimit[count2][3]*100));
    }
    sprintf(tempstring,"%s\n$$VOLTALERTLO %s\n$$VOLTALERTHI %s\n$$VOLTSHUTDOWNLO %s\n$$VOLTSHUTDOWNHI %s\n",
            tempstring,voltlow,volthigh,voltcritlow,voltcrithigh); 
    sprintf(tempstring,"%s$$TEMPALERTLO %s\n$$TEMPALERTHI %s\n$$TEMPSHUTDOWNLO %s\n$$TEMPSHUTDOWNHI %s\n",
            tempstring,templow,temphigh,tempcritlow,tempcrithigh); 
    sprintf(tempstring,"%s$$FANALERTLO %s\n$$FANALERTHI %s\n$$FANSHUTDOWNLO %s\n$$FANSHUTDOWNHI %s\n",
            tempstring,fanlow,fanhigh,fancritlow,fancrithigh); 
    fprintf(outputfile,"%s",tempstring);
  } 
  fclose(outputfile);
  umask(oldumask);
  chown(autonodeconfdat,-1,apachegid);
}

void cluster::outputdata() {
  int count, count2;
  FILE *outputfile;
  char tempstring[1000];

  outputfile=fopen(nodewatchdat,"w");
  if(outputfile==NULL) {
    fprintf(stderr,"Could not open %s for writing.\n",nodewatchdat);
    exit(1);
  }

  for (count=0; count<numnodes ; count++){
    sprintf(tempstring,"$TIME %x\n$NODE 1 %s\n$$DEAD %i\n$$FAN",
            (unsigned int)mynodes[count].lastupdate, mynodes[count].name, mynodes[count].isdead());
    for (count2=0; count2<mynodes[count].thischassis->numfans; count2++) {
      sprintf(tempstring,"%s %i", tempstring, (int)mynodes[count].fan[count2]);
    }
    sprintf(tempstring,"%s\n$$TEMP", tempstring);
    for (count2=0; count2<mynodes[count].thischassis->numtemps; count2++) {
      sprintf(tempstring,"%s %i", tempstring, (int)(mynodes[count].temp[count2]*100));
    }
    sprintf(tempstring,"%s\n$$VOLT", tempstring);
    for (count2=0; count2<mynodes[count].thischassis->numvolts; count2++) {
      sprintf(tempstring,"%s %i", tempstring, (int)(mynodes[count].volt[count2]*100));
    }
    sprintf(tempstring,"%s\n$$POWER %i",tempstring,mynodes[count].ispowered());
    fprintf(outputfile,"%s\n",tempstring);
  }
  fclose(outputfile);
}

chassis *cluster::findchassis(char *name) {
  for(int i=0;i<numchassis;i++){
    if (strcmp(mychassis[i].name,name)==0) return &mychassis[i];
  }
  return NULL;
}

node *cluster::findnode(char *name) {
  for(int i=0;i<numnodes;i++){
    if (strcmp(mynodes[i].name,name)==0) return &mynodes[i];
  }
  return NULL;
}

void cluster::getdata() {
  for(int i=0;i<numnodes;i++){
    mynodes[i].getdata();
  }
}

void cluster::checklimits() {
  for(int i=0;i<numnodes;i++){
    mynodes[i].checklimits();
  }
}

bool cluster::poweroffnode(char *name) {
  for(int i=0; i<numnodes; i++){
    if (strcmp(mynodes[i].name, name)==0) {
      mynodes[i].shutdown();
      return true;
    }
  }
  return false;
}

bool cluster::registersensor(char *name, ipmi_sensor_t *sensor) {
  for(int i=0; i<numnodes; i++){
    if (strcmp(mynodes[i].name, name)==0) {
      return mynodes[i].registersensor(sensor);
    }
  }
  return false;
} 

bool cluster::reportsensor(char *name, char *sensorname, double value) {
  for(int i=0; i<numnodes; i++){
    if (strcmp(mynodes[i].name, name)==0) {
      return mynodes[i].reportsensor(sensorname,value);
    }
  }
  return false;
} 


bool cluster::zerosensors(char *name) {
  for(int i=0; i<numnodes; i++){
    if (strcmp(mynodes[i].name, name)==0) {
      return mynodes[i].zerosensors();
    }
  }
  return false;
}


bool cluster::reportdomain(char *name, int err) {
  for(int i=0; i<numnodes; i++){
    if (strcmp(mynodes[i].name, name)==0) {
      return mynodes[i].reportdomain(err);
    }
  }
  return false;
}
